@@ -111,6 +111,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	 *    START_ARRAY
 	 *
 	 */
+	int clean = 0;
 	int must_close = 0;
 	int old_linux = 0;
 	int vers = 0; /* Keep gcc quite - it really is initialised */
@@ -135,6 +136,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	int chosen_drive;
 	int change = 0;
 	int inargv = 0;
+	int bitmap_done;
 	int start_partial_ok = (runstop >= 0) && (force || devlist==NULL || mdfd < 0);
 	unsigned int num_devs;
 	mddev_dev_t tmpdev;
@@ -412,6 +414,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	}
 
 	/* Ok, no bad inconsistancy, we can try updating etc */
+	bitmap_done = 0;
 	for (tmpdev = devlist; tmpdev; tmpdev=tmpdev->next) if (tmpdev->used == 1) {
 		char *devname = tmpdev->devname;
 		struct stat stb;
@@ -468,10 +471,12 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 				close(dfd);
 
 			if (strcmp(update, "uuid")==0 &&
-			    ident->bitmap_fd)
-				if (bitmap_update_uuid(ident->bitmap_fd, info.uuid) != 0)
-					fprintf(stderr, Name ": Could not update uuid on %s.\n",
-						devname);
+			    ident->bitmap_fd >= 0 && !bitmap_done) {
+				if (bitmap_update_uuid(ident->bitmap_fd, info.uuid, st->ss->swapuuid) != 0)
+					fprintf(stderr, Name ": Could not update uuid on external bitmap.\n");
+				else
+					bitmap_done = 1;
+			}
 		} else
 #endif
 		{
@@ -583,6 +588,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	}
 
 	st->ss->getinfo_super(&info, first_super);
+	clean = info.array.state & 1;
 
 	/* now we have some devices that might be suitable.
 	 * I wonder how many
@@ -617,13 +623,14 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		}
 	}
 	while (force && !enough(info.array.level, info.array.raid_disks,
-				info.array.layout,
+				info.array.layout, 1,
 				avail, okcnt)) {
 		/* Choose the newest best drive which is
 		 * not up-to-date, update the superblock
 		 * and add it.
 		 */
 		int fd;
+		long long current_events;
 		chosen_drive = -1;
 		for (i=0; i<info.array.raid_disks && i < bestcnt; i++) {
 			int j = best[i];
@@ -636,6 +643,8 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		}
 		if (chosen_drive < 0)
 			break;
+		current_events = devices[chosen_drive].events;
+	add_another:
 		if (verbose >= 0)
 			fprintf(stderr, Name ": forcing event count in %s(%d) from %d upto %d\n",
 				devices[chosen_drive].devname, devices[chosen_drive].raid_disk,
@@ -656,7 +665,9 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			continue;
 		}
 		info.events = devices[most_recent].events;
-		st->ss->update_super(&info, super, "force", devices[chosen_drive].devname, verbose, 0, NULL);
+		st->ss->update_super(&info, super, "force-one",
+				     devices[chosen_drive].devname, verbose,
+				     0, NULL);
 
 		if (st->ss->store_super(st, fd, super)) {
 			close(fd);
@@ -672,6 +683,20 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		avail[chosen_drive] = 1;
 		okcnt++;
 		free(super);
+
+		/* If there are any other drives of the same vintage,
+		 * add them in as well.  We can't lose and we might gain
+		 */
+		for (i=0; i<info.array.raid_disks && i < bestcnt ; i++) {
+			int j = best[i];
+			if (j >= 0 &&
+			    !devices[j].uptodate &&
+			    devices[j].events > 0 &&
+			    devices[j].events == current_events) {
+				chosen_drive = j;
+				goto add_another;
+			}
+		}
 	}
 
 	/* Now we want to look at the superblock which the kernel will base things on
@@ -752,10 +777,14 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		}
 #endif
 	}
-	if (force && okcnt == info.array.raid_disks-1) {
-		/* FIXME check event count */
-		change += st->ss->update_super(&info, super, "force",
-					devices[chosen_drive].devname, verbose, 0, NULL);
+	if (force && !clean &&
+	    !enough(info.array.level, info.array.raid_disks,
+		    info.array.layout, clean,
+		    avail, okcnt)) {
+		change += st->ss->update_super(&info, super, "force-array",
+					devices[chosen_drive].devname, verbose,
+					       0, NULL);
+		clean = 1;
 	}
 
 	if (change) {
@@ -892,7 +921,8 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		
 		if (runstop == 1 ||
 		    (runstop <= 0 &&
-		     ( enough(info.array.level, info.array.raid_disks, info.array.layout, avail, okcnt) &&
+		     ( enough(info.array.level, info.array.raid_disks,
+			      info.array.layout, clean, avail, okcnt) &&
 		       (okcnt >= req_cnt || start_partial_ok)
 			     ))) {
 			if (ioctl(mdfd, RUN_ARRAY, NULL)==0) {
@@ -921,8 +951,8 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 					while (usecs < 1000) {
 						mdfd = open(mddev, O_RDONLY);
 						if (mdfd >= 0) {
-							unsigned long size;
-							if (ioctl(mdfd, BLKGETSIZE, &size) == 0 &&
+							unsigned long long size;
+							if (get_dev_size(mdfd, NULL, &size) &&
 							    size > 0)
 								break;
 							close(mdfd);
@@ -935,6 +965,19 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			}
 			fprintf(stderr, Name ": failed to RUN_ARRAY %s: %s\n",
 				mddev, strerror(errno));
+
+			if (!enough(info.array.level, info.array.raid_disks,
+				    info.array.layout, 1, avail, okcnt))
+				fprintf(stderr, Name ": Not enough devices to "
+					"start the array.\n");
+			else if (!enough(info.array.level,
+					 info.array.raid_disks,
+					 info.array.layout, clean,
+					 avail, okcnt))
+				fprintf(stderr, Name ": Not enough devices to "
+					"start the array while not clean "
+					"- consider --force.\n");
+
 			if (must_close) close(mdfd);
 			return 1;
 		}
@@ -947,12 +990,20 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			if (must_close) close(mdfd);
 			return 0;
 		}
-		if (verbose >= 0) {
+		if (verbose >= -1) {
 			fprintf(stderr, Name ": %s assembled from %d drive%s", mddev, okcnt, okcnt==1?"":"s");
 			if (sparecnt)
 				fprintf(stderr, " and %d spare%s", sparecnt, sparecnt==1?"":"s");
-			if (!enough(info.array.level, info.array.raid_disks, info.array.layout, avail, okcnt))
+			if (!enough(info.array.level, info.array.raid_disks,
+				    info.array.layout, 1, avail, okcnt))
 				fprintf(stderr, " - not enough to start the array.\n");
+			else if (!enough(info.array.level,
+					 info.array.raid_disks,
+					 info.array.layout, clean,
+					 avail, okcnt))
+				fprintf(stderr, " - not enough to start the "
+					"array while not clean - consider "
+					"--force.\n");
 			else {
 				if (req_cnt == info.array.raid_disks)
 					fprintf(stderr, " - need all %d to start it", req_cnt);
