@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -33,8 +33,10 @@
 #include	<sys/wait.h>
 #include	<sys/signal.h>
 #include	<values.h>
+#include	<syslog.h>
 
-static void alert(char *event, char *dev, char *disc, char *mailaddr, char *cmd);
+static void alert(char *event, char *dev, char *disc, char *mailaddr, char *mailfrom,
+		  char *cmd, int dosyslog);
 
 static char *percentalerts[] = { 
 	"RebuildStarted",
@@ -44,10 +46,15 @@ static char *percentalerts[] = {
 	"Rebuild80",
 };
 
+/* The largest number of disks current arrays can manage is 384
+ * This really should be dynamically, but that will have to wait
+ * At least it isn't MD_SB_DISKS.
+ */
+#define MaxDisks 384
 int Monitor(mddev_dev_t devlist,
 	    char *mailaddr, char *alert_cmd,
 	    int period, int daemonise, int scan, int oneshot,
-	    char *config, int test)
+	    int dosyslog, int test, char* pidfile)
 {
 	/*
 	 * Every few seconds, scan every md device looking for changes
@@ -99,22 +106,25 @@ int Monitor(mddev_dev_t devlist,
 		char *spare_group;
 		int active, working, failed, spare, raid;
 		int expected_spares;
-		int devstate[MD_SB_DISKS];
-		int devid[MD_SB_DISKS];
+		int devstate[MaxDisks];
+		int devid[MaxDisks];
 		int percent;
 		struct state *next;
 	} *statelist = NULL;
 	int finished = 0;
 	struct mdstat_ent *mdstat = NULL;
+	char *mailfrom = NULL;
 
 	if (!mailaddr) {
-		mailaddr = conf_get_mailaddr(config);
+		mailaddr = conf_get_mailaddr();
 		if (mailaddr && ! scan)
 			fprintf(stderr, Name ": Monitor using email address \"%s\" from config file\n",
 			       mailaddr);
 	}
+	mailfrom = conf_get_mailfrom();
+
 	if (!alert_cmd) {
-		alert_cmd = conf_get_program(config);
+		alert_cmd = conf_get_program();
 		if (alert_cmd && ! scan)
 			fprintf(stderr, Name ": Monitor using program \"%s\" from config file\n",
 			       alert_cmd);
@@ -127,7 +137,18 @@ int Monitor(mddev_dev_t devlist,
 	if (daemonise) {
 		int pid = fork();
 		if (pid > 0) {
-			printf("%d\n", pid);
+			if (!pidfile)
+				printf("%d\n", pid);
+			else {
+				FILE *pid_file;
+				pid_file=fopen(pidfile, "w");
+				if (!pid_file)
+					perror("cannot create pid file");
+				else {
+					fprintf(pid_file,"%d\n", pid);
+					fclose(pid_file);
+				}
+			}
 			return 0;
 		}
 		if (pid < 0) {
@@ -142,7 +163,7 @@ int Monitor(mddev_dev_t devlist,
 	}
 
 	if (devlist == NULL) {
-		mddev_ident_t mdlist = conf_get_ident(config, NULL);
+		mddev_ident_t mdlist = conf_get_ident(NULL);
 		for (; mdlist; mdlist=mdlist->next) {
 			struct state *st = malloc(sizeof *st);
 			if (st == NULL)
@@ -163,6 +184,7 @@ int Monitor(mddev_dev_t devlist,
 	} else {
 		mddev_dev_t dv;
 		for (dv=devlist ; dv; dv=dv->next) {
+			mddev_ident_t mdlist = conf_get_ident(dv->devname);
 			struct state *st = malloc(sizeof *st);
 			if (st == NULL)
 				continue;
@@ -174,6 +196,11 @@ int Monitor(mddev_dev_t devlist,
 			st->percent = -2;
 			st->expected_spares = -1;
 			st->spare_group = NULL;
+			if (mdlist) {
+				st->expected_spares = mdlist->spare_disks;
+				if (mdlist->spare_group)
+					st->spare_group = strdup(mdlist->spare_group);
+			}
 			statelist = st;
 		}
 	}
@@ -185,22 +212,23 @@ int Monitor(mddev_dev_t devlist,
 
 		if (mdstat)
 			free_mdstat(mdstat);
-		mdstat = mdstat_read();
+		mdstat = mdstat_read(oneshot?0:1, 0);
 
 		for (st=statelist; st; st=st->next) {
+			struct { int state, major, minor; } info[MaxDisks];
 			mdu_array_info_t array;
-			struct mdstat_ent *mse;
+			struct mdstat_ent *mse = NULL, *mse2;
 			char *dev = st->devname;
 			int fd;
 			unsigned int i;
 
 			if (test)
-				alert("TestMessage", dev, NULL, mailaddr, alert_cmd);
+				alert("TestMessage", dev, NULL, mailaddr, mailfrom, alert_cmd, dosyslog);
 			fd = open(dev, O_RDONLY);
 			if (fd < 0) {
 				if (!st->err)
 					alert("DeviceDisappeared", dev, NULL,
-					      mailaddr, alert_cmd);
+					      mailaddr, mailfrom, alert_cmd, dosyslog);
 /*					fprintf(stderr, Name ": cannot open %s: %s\n",
 						dev, strerror(errno));
 */				st->err=1;
@@ -209,17 +237,18 @@ int Monitor(mddev_dev_t devlist,
 			if (ioctl(fd, GET_ARRAY_INFO, &array)<0) {
 				if (!st->err)
 					alert("DeviceDisappeared", dev, NULL,
-					      mailaddr, alert_cmd);
+					      mailaddr, mailfrom, alert_cmd, dosyslog);
 /*					fprintf(stderr, Name ": cannot get array info for %s: %s\n",
 						dev, strerror(errno));
 */				st->err=1;
 				close(fd);
 				continue;
 			}
-			if (array.level != 1 && array.level != 5 && array.level != -4) {
+			if (array.level != 1 && array.level != 5 && array.level != -4 &&
+				array.level != 6 && array.level != 10) {
 				if (!st->err)
 					alert("DeviceDisappeared", dev, "Wrong-Level",
-					      mailaddr, alert_cmd);
+					      mailaddr, mailfrom, alert_cmd, dosyslog);
 				st->err = 1;
 				close(fd);
 				continue;
@@ -228,16 +257,18 @@ int Monitor(mddev_dev_t devlist,
 				struct stat stb;
 				if (fstat(fd, &stb) == 0 &&
 				    (S_IFMT&stb.st_mode)==S_IFBLK) {
-					if (MINOR(stb.st_rdev) == 9)
-						st->devnum = MINOR(stb.st_rdev);
+					if (major(stb.st_rdev) == MD_MAJOR)
+						st->devnum = minor(stb.st_rdev);
 					else
-						st->devnum = -1- (MINOR(stb.st_rdev)>>6);
+						st->devnum = -1- (minor(stb.st_rdev)>>6);
 				}
 			}
 
-			for (mse = mdstat ; mse ; mse=mse->next)
-				if (mse->devnum == st->devnum)
-					mse->devnum = MAXINT; /* flag it as "used" */
+			for (mse2 = mdstat ; mse2 ; mse2=mse2->next)
+				if (mse2->devnum == st->devnum) {
+					mse2->devnum = MAXINT; /* flag it as "used" */
+					mse = mse2;
+				}
 
 			if (st->utime == array.utime &&
 			    st->failed == array.failed_disks &&
@@ -254,45 +285,83 @@ int Monitor(mddev_dev_t devlist,
 			    mse &&	/* is in /proc/mdstat */
 			    mse->pattern && strchr(mse->pattern, '_') /* degraded */
 				)
-				alert("DegradedArray", dev, NULL, mailaddr, alert_cmd);
+				alert("DegradedArray", dev, NULL, mailaddr, mailfrom, alert_cmd, dosyslog);
 
 			if (st->utime == 0 && /* new array */
 			    st->expected_spares > 0 && 
 			    array.spare_disks < st->expected_spares) 
-				alert("SparesMissing", dev, NULL, mailaddr, alert_cmd);
+				alert("SparesMissing", dev, NULL, mailaddr, mailfrom, alert_cmd, dosyslog);
 			if (mse &&
 			    st->percent == -1 && 
 			    mse->percent >= 0)
-				alert("RebuildStarted", dev, NULL, mailaddr, alert_cmd);
+				alert("RebuildStarted", dev, NULL, mailaddr, mailfrom, alert_cmd, dosyslog);
 			if (mse &&
 			    st->percent >= 0 &&
 			    mse->percent >= 0 &&
 			    (mse->percent / 20) > (st->percent / 20))
 				alert(percentalerts[mse->percent/20],
-				      dev, NULL, mailaddr, alert_cmd);
+				      dev, NULL, mailaddr, mailfrom, alert_cmd, dosyslog);
 
 			if (mse &&
 			    mse->percent == -1 &&
-			    st->percent >= 0)
-				alert("RebuildFinished", dev, NULL, mailaddr, alert_cmd);
+			    st->percent >= 0) {
+				/* Rebuild/sync/whatever just finished.
+				 * If there is a number in /mismatch_cnt,
+				 * we should report that.
+				 */
+				struct sysarray *sra =
+				       sysfs_read(-1, st->devnum, GET_MISMATCH);
+				if (sra && sra->mismatch_cnt > 0) {
+					char cnt[40];
+					sprintf(cnt, " mismatches found: %d", sra->mismatch_cnt);
+					alert("RebuildFinished", dev, cnt, mailaddr, mailfrom, alert_cmd, dosyslog);
+				} else
+					alert("RebuildFinished", dev, NULL, mailaddr, mailfrom, alert_cmd, dosyslog);
+				if (sra)
+					free(sra);
+			}
 
 			if (mse)
 				st->percent = mse->percent;
-					
-			for (i=0; i<MD_SB_DISKS; i++) {
+
+
+			for (i=0; i<MaxDisks && i <= array.raid_disks + array.nr_disks;
+			     i++) {
 				mdu_disk_info_t disc;
+				if (ioctl(fd, GET_DISK_INFO, &disc) >= 0) {
+					info[i].state = disc.state;
+					info[i].major = disc.major;
+					info[i].minor = disc.minor;
+				} else
+					info[i].major = info[i].minor = 0;
+			}
+			close(fd);
+
+			for (i=0; i<MaxDisks; i++) {
+				mdu_disk_info_t disc = {0};
 				int newstate=0;
 				int change;
 				char *dv = NULL;
 				disc.number = i;
-				if (ioctl(fd, GET_DISK_INFO, &disc)>= 0) {
-					newstate = disc.state;
-					dv = map_dev(disc.major, disc.minor);
-				} else if (mse &&  mse->pattern && i < strlen(mse->pattern))
+				if (i > array.raid_disks + array.nr_disks) {
+					newstate = 0;
+					disc.major = disc.minor = 0;
+				} else if (info[i].major || info[i].minor) {
+					newstate = info[i].state;
+					dv = map_dev(info[i].major, info[i].minor, 1);
+					disc.state = newstate;
+					disc.major = info[i].major;
+					disc.minor = info[i].minor;
+				} else if (mse &&  mse->pattern && i < strlen(mse->pattern)) {
 					switch(mse->pattern[i]) {
 					case 'U': newstate = 6 /* ACTIVE/SYNC */; break;
 					case '_': newstate = 0; break;
 					}
+					disc.major = disc.minor = 0;
+				}
+				if (dv == NULL && st->devid[i])
+					dv = map_dev(major(st->devid[i]),
+						     minor(st->devid[i]), 1);
 				change = newstate ^ st->devstate[i];
 				if (st->utime && change && !st->err) {
 					if (i < (unsigned)array.raid_disks &&
@@ -300,24 +369,23 @@ int Monitor(mddev_dev_t devlist,
 					     ((st->devstate[i]&change)&(1<<MD_DISK_ACTIVE)) ||
 					     ((st->devstate[i]&change)&(1<<MD_DISK_SYNC)))
 						)
-						alert("Fail", dev, dv, mailaddr, alert_cmd);
+						alert("Fail", dev, dv, mailaddr, mailfrom, alert_cmd, dosyslog);
 					else if (i >= (unsigned)array.raid_disks &&
 						 (disc.major || disc.minor) &&
-						 st->devid[i] == MKDEV(disc.major, disc.minor) &&
+						 st->devid[i] == makedev(disc.major, disc.minor) &&
 						 ((newstate&change)&(1<<MD_DISK_FAULTY))
 						)
-						alert("FailSpare", dev, dv, mailaddr, alert_cmd);
+						alert("FailSpare", dev, dv, mailaddr, mailfrom, alert_cmd, dosyslog);
 					else if (i < (unsigned)array.raid_disks &&
 						 (((st->devstate[i]&change)&(1<<MD_DISK_FAULTY)) ||
 						  ((newstate&change)&(1<<MD_DISK_ACTIVE)) ||
 						  ((newstate&change)&(1<<MD_DISK_SYNC)))
 						)
-						alert("SpareActive", dev, dv, mailaddr, alert_cmd);
+						alert("SpareActive", dev, dv, mailaddr, mailfrom, alert_cmd, dosyslog);
 				}
 				st->devstate[i] = disc.state;
-				st->devid[i] = MKDEV(disc.major, disc.minor);
+				st->devid[i] = makedev(disc.major, disc.minor);
 			}
-			close(fd);
 			st->active = array.active_disks;
 			st->working = array.working_disks;
 			st->spare = array.spare_disks;
@@ -345,10 +413,12 @@ int Monitor(mddev_dev_t devlist,
 					    ioctl(fd, GET_ARRAY_INFO, &array)< 0) {
 						/* no such array */
 						if (fd >=0) close(fd);
+						put_md_name(st->devname);
 						free(st->devname);
 						free(st);
 						continue;
 					}
+					close(fd);
 					st->utime = 0;
 					st->next = statelist;
 					st->err = 1;
@@ -357,7 +427,7 @@ int Monitor(mddev_dev_t devlist,
 					st->spare_group = NULL;
 					st->expected_spares = -1;
 					statelist = st;
-					alert("NewArray", st->devname, NULL, mailaddr, alert_cmd);
+					alert("NewArray", st->devname, NULL, mailaddr, mailfrom, alert_cmd, dosyslog);
 					new_found = 1;
 				}
 		}
@@ -386,7 +456,7 @@ int Monitor(mddev_dev_t devlist,
 							if (fd2>=0) close(fd2);
 							continue;
 						}
-						for (d=st2->raid; d<MD_SB_DISKS; d++) {
+						for (d=st2->raid; d < MaxDisks; d++) {
 							if (st2->devid[d] > 0 &&
 							    st2->devstate[d] == 0) {
 								dev = st2->devid[d];
@@ -398,7 +468,7 @@ int Monitor(mddev_dev_t devlist,
 								  (unsigned long)dev) == 0) {
 								if (ioctl(fd1, HOT_ADD_DISK,
 									  (unsigned long)dev) == 0) {
-									alert("MoveSpare", st->devname, st2->devname, mailaddr, alert_cmd);
+									alert("MoveSpare", st->devname, st2->devname, mailaddr, mailfrom, alert_cmd, dosyslog);
 									close(fd1);
 									close(fd2);
 									break;
@@ -414,16 +484,21 @@ int Monitor(mddev_dev_t devlist,
 			if (oneshot)
 				break;
 			else
-				sleep(period);
+				mdstat_wait(period);
 		}
 		test = 0;
 	}
+	if (pidfile)
+		unlink(pidfile);
 	return 0;
 }
 
 
-static void alert(char *event, char *dev, char *disc, char *mailaddr, char *cmd)
+static void alert(char *event, char *dev, char *disc, char *mailaddr, char *mailfrom, char *cmd,
+		  int dosyslog)
 {
+	int priority;
+
 	if (!cmd && !mailaddr) {
 		time_t now = time(0);
 	       
@@ -445,13 +520,18 @@ static void alert(char *event, char *dev, char *disc, char *mailaddr, char *cmd)
 	if (mailaddr && 
 	    (strncmp(event, "Fail", 4)==0 || 
 	     strncmp(event, "Test", 4)==0 ||
+	     strncmp(event, "Spares", 6)==0 ||
 	     strncmp(event, "Degrade", 7)==0)) {
 		FILE *mp = popen(Sendmail, "w");
 		if (mp) {
+			FILE *mdstat;
 			char hname[256];
 			gethostname(hname, sizeof(hname));
 			signal(SIGPIPE, SIG_IGN);
-			fprintf(mp, "From: " Name " monitoring <root>\n");
+			if (mailfrom)
+				fprintf(mp, "From: %s\n", mailfrom);
+			else
+				fprintf(mp, "From: " Name " monitoring <root>\n");
 			fprintf(mp, "To: %s\n", mailaddr);
 			fprintf(mp, "Subject: %s event on %s:%s\n\n", event, dev, hname);
 
@@ -460,13 +540,83 @@ static void alert(char *event, char *dev, char *disc, char *mailaddr, char *cmd)
 
 			fprintf(mp, "A %s event had been detected on md device %s.\n\n", event, dev);
 
-			if (disc)
+			if (disc && disc[0] != ' ')
 				fprintf(mp, "It could be related to component device %s.\n\n", disc);
+			if (disc && disc[0] == ' ')
+				fprintf(mp, "Extra information:%s.\n\n", disc);
 
 			fprintf(mp, "Faithfully yours, etc.\n");
+
+			mdstat = fopen("/proc/mdstat", "r");
+			if (mdstat) {
+				char buf[8192];
+				int n;
+				fprintf(mp, "\nP.S. The /proc/mdstat file currently contains the following:\n\n");
+				while ( (n=fread(buf, 1, sizeof(buf), mdstat)) > 0)
+					n=fwrite(buf, 1, n, mp); /* yes, i don't care about the result */
+				fclose(mdstat);
+			}
 			fclose(mp);
 		}
 
 	}
-	/* FIXME log the event to syslog maybe */
+
+	/* log the event to syslog maybe */
+	if (dosyslog) {
+		/* Log at a different severity depending on the event.
+		 *
+		 * These are the critical events:  */
+		if (strncmp(event, "Fail", 4)==0 ||
+		    strncmp(event, "Degrade", 7)==0 ||
+		    strncmp(event, "DeviceDisappeared", 17)==0)
+			priority = LOG_CRIT;
+		/* Good to know about, but are not failures: */
+		else if (strncmp(event, "Rebuild", 7)==0 ||
+			 strncmp(event, "MoveSpare", 9)==0 ||
+			 strncmp(event, "Spares", 6) != 0)
+			priority = LOG_WARNING;
+		/* Everything else: */
+		else
+			priority = LOG_INFO;
+
+		if (disc)
+			syslog(priority, "%s event detected on md device %s, component device %s", event, dev, disc);
+		else
+			syslog(priority, "%s event detected on md device %s", event, dev);
+	}
+}
+
+/* Not really Monitor but ... */
+int Wait(char *dev)
+{
+	struct stat stb;
+	int devnum;
+	int rv = 1;
+
+	if (stat(dev, &stb) != 0) {
+		fprintf(stderr, Name ": Cannot find %s: %s\n", dev,
+			strerror(errno));
+		return 2;
+	}
+	if (major(stb.st_rdev) == MD_MAJOR)
+		devnum = minor(stb.st_rdev);
+	else
+		devnum = -1-(minor(stb.st_rdev)/64);
+
+	while(1) {
+		struct mdstat_ent *ms = mdstat_read(1, 0);
+		struct mdstat_ent *e;
+
+		for (e=ms ; e; e=e->next)
+			if (e->devnum == devnum)
+				break;
+
+		if (!e || e->percent < 0) {
+			free_mdstat(ms);
+			return rv;
+		}
+		free(ms);
+		rv = 0;
+		mdstat_wait(5);
+	}
 }
